/*********************************************************************
 * Desktop Station AUTO Sketch for DS shield series
 *
 * Copyright (C) 2017 Yaasan
 *
 */

// include the library code:
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM.h"
#include "Functions.h"
#include "TrackReporterS88_DS.h"

#define MAX_S88DECODER 1

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2

#define PIN_RUNLED	2

#define MODE_SPEEDSTEP128 2

#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_OV  614  // 1024d * (1.2 / 8) * 20V / 5V


/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);

/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;

//Task Schedule
unsigned long gPreviousL8 = 0; // 1000ms interval


//コマンド関数定義
void CMD_Wait(word inWaitMs);
void CMD_Power(byte inState);
void CMD_LocSpeed(word inAddress, word inSpeed);
void CMD_LocDirection(word inAddress, byte inDirection);
void CMD_LocFunction(word inAddress, byte inFunctionNo, byte inFunctionState);
void CMD_LocFunctionMomentary(word inAddress, byte inFunctionNo);
void CMD_Turnout(word inAddress, byte inDirection);
void CMD_ScanS88(void);
byte CMD_GetS88(byte inNo);
void CMD_END(void);

/***********************************************************************
* 
* Boot setup
* 
************************************************************************/

void setup()
{
	cli();
	
	//タイマ0を強制的に戻すr
	TCCR0B = (TCCR0B & 0b11111000) | 0x03;
	//Timer1 init
	TCCR1A = 0;
	TCCR1B = 0;
	TCNT1  = 0;

	OCR1A = 20;            // 1Hz = 62500d, 100Hz = 625d
	TCCR1B |= (1 << WGM12);   // CTC mode
	TCCR1B |= (1 << CS12);    // 256 prescaler 
	TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt

	DSCore.Init();

	//Pin mode
	pinMode(A1, INPUT);//Edc
	pinMode(PIN_RUNLED, OUTPUT);
	
	/* LED turn on */
	changeLED(0);
	
	/* Clear data of the S88 decoder */
	reporter.refresh();
	
	/* Debug message for PC */
	Serial.begin(115200);
	Serial.println("--------------------------------------");
	Serial.println("Desktop Station AUTO R1d 2017/11/12  ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");
	
	//Reset task
	gPreviousL8 = millis();
	
	
	
	sei();    //ENABLE INTERRUPTION
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	//User initial
	user_init();
	
}

/***********************************************************************
* 
* Command for locomotives and turnouts
* 
************************************************************************/

void CMD_Wait(word inWaitMs)
{
	unsigned long aTemp = millis() + (inWaitMs >> 2);
	
	//なんか時間が合わない。Timer0のプリスケーラがおかしい？
	while ( aTemp > millis() )
	{
		//NOP
		__asm__("nop\n\t");
	}
	
}

void CMD_Power(byte inState)
{
	/* LED状態の変更 */
	changeLED(inState);
	
	DSCore.SetPower(inState);
}

void CMD_LocSpeed(word inAddress, word inSpeed)
{
	byte aMode = MODE_SPEEDSTEP128;
	
	if( inAddress <= 256)
	{
		aMode = 0;
	}
	
	
	DSCore.SetLocoSpeedEx(inAddress, inSpeed, aMode);
}

void CMD_LocDirection(word inAddress, byte inDirection)
{
	DSCore.SetLocoDirection(inAddress, inDirection);
}

void CMD_LocFunction(word inAddress, byte inFunctionNo, byte inFunctionState)
{
	DSCore.SetLocoFunction(inAddress, inFunctionNo + 1, inFunctionState);
}

void CMD_LocFunctionMomentary(word inAddress, byte inFunctionNo)
{
	DSCore.SetLocoFunction(inAddress, inFunctionNo + 1, 1);
	CMD_Wait(200);
	DSCore.SetLocoFunction(inAddress, inFunctionNo + 1, 0);
}


void CMD_Turnout(word inAddress, byte inDirection)
{
	DSCore.SetTurnout(inAddress - 1, inDirection);
}

void CMD_ScanS88(void)
{
	reporter.refresh(MAX_S88DECODER);
}

byte CMD_GetS88(byte inNo)
{
	word aFlags = reporter.getByte(0) + reporter.getByte(1);
	
	return (aFlags >> inNo) & 1;
}

void CMD_END(void)
{
	/* 永久ループ（終了） */
	while(1){}
}


/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

ISR(TIMER1_COMPA_vect)
{
	uint8_t aErrNo = 0;
	word aEdc;
	aErrNo = 0;
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	if( (millis() - gPreviousL8) >= 200)
	{
		//Reset task
		gPreviousL8 = millis();
		
		//LED Control
		taskLED();
		
		/* Sensor check */
		aEdc = analogRead(A1);
		
		/* OV check */
		if( aEdc > THRESHOLD_OV )
		{
			aErrNo = 2;
		}
		/* LV check */
		if( aEdc < THRESHOLD_EDC )
		{
			aErrNo = 1;
			
		}
		/* エラー処理 */
		if( aErrNo > 0)
		{
			if( DSCore.IsPower() == 1)
			{
				PowerOffByErr(aErrNo);
			}
			
		}
	}
	
}

/***********************************************************************
* 
* Main Loop
* 
************************************************************************/

void loop()
{
	
	user_program();
	
}

/***********************************************************************
* 
* RUN LED Control functions
* 
************************************************************************/

void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}


void changeLED(byte inState)
{
	if( inState == 1)
	{
		/* 状態変更 */
		gLED_State = LEDSTATE_RUN;
	}
	else
	{
		/* 状態変更 */
		gLED_State = LEDSTATE_STOP;
	}
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, 0);
		
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, 1);
		break;
		
	case LEDSTATE_ERR:
		break;
	}
}

void taskLED(void)
{
	
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		digitalWrite(PIN_RUNLED, gLED_Toggle);
		gLED_Toggle = (~gLED_Toggle) & 1;
		
		break;
	}
}

/*****************************************************************/
/*  User Program Area                                            */
/*****************************************************************/

void user_init(void)
{
	
	pinMode(8, INPUT);
	pinMode(11, INPUT);
	pinMode(12, INPUT);
	
	digitalWrite(8, HIGH);//内部プルアップON
	digitalWrite(11, HIGH);//内部プルアップON
	digitalWrite(12, HIGH);//内部プルアップON
	
	//F0 Light ON
	CMD_LocFunction(ADDR_DCC + 3, 0, 1);
	
	//SPEED 0
	CMD_LocSpeed(ADDR_DCC + 3, 0);
	
	//POWER ON
//	CMD_Power(1);
//	Serial.println("Power On");
}


void user_program(void)
{
	static byte sButtonState = 1;
  CMD_Wait(20000);	
	
	//ボタンが押されたとき、digitalReadは0を返す。押さないときは1を返す。
	
	if( digitalRead(8) == 0)
       {
       //POWER ON
       CMD_Power(1);
       Serial.println("Power On");
       }
   if( digitalRead(8) == 1)
	 {
		      Serial.println("Button 1");

		     sButtonState = 1;
		     CMD_Wait(100);
		
		     /* 前進 */
		     //FWD
		     CMD_LocDirection(ADDR_DCC + 3, 0);
		     CMD_Wait(2000);
   
		     //警笛
		     CMD_LocFunction(ADDR_DCC + 3, 11, 1);
		     CMD_Wait(7000);
		     //CMD_LocFunction(ADDR_DCC + 3, 11, 0);
         //CMD_Wait(2000);

                //繰り返し。車両のスピードを変える（加速）

                for( int i = 0; i < 170; i++)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

		      //急加速
		      //CMD_LocSpeed(ADDR_DCC + 3, 100);

                //定速時間
                CMD_Wait(20000);    //定速時間20秒

                //警笛
                CMD_LocFunction(ADDR_DCC + 3, 11, 0);
                CMD_Wait(2000);
                
                //定速時間
                CMD_Wait(20000);    //定速時間20秒

                //繰り返し。車両のスピードを変える(減速1）
                for( int i = 170; i > 32; i--)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

                //絶気モード
                //CMD_LocFunction(ADDR_DCC + 3, 3, 1);
                //CMD_Wait(2000);
                
                //繰り返し。車両のスピードを変える(減速2）
                for( int i = 32; i > 0; i--)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

                //停止時間
                CMD_LocSpeed(ADDR_DCC + 3, 0);
                CMD_Wait(2000);
                
                //絶気モード解除
                //CMD_LocFunction(ADDR_DCC + 3, 3, 0);
                //CMD_Wait(2000);
                
                CMD_Wait(40000);   //停止時間40秒

		      /* 後進 */
		     //REV
		     CMD_LocDirection(ADDR_DCC + 3, 1);
		     CMD_Wait(2000);
   
		     //警笛
		     CMD_LocFunction(ADDR_DCC + 3, 11, 1);
		     CMD_Wait(7000);
		     //CMD_LocFunction(ADDR_DCC + 3, 11, 0);
         //CMD_Wait(2000);

                //繰り返し。車両のスピードを変える（加速）

                for( int i = 0; i < 170; i++)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

		      //急加速
		      //CMD_LocSpeed(ADDR_DCC + 3, 100);

                //定速時間
                CMD_Wait(20000);    //定速時間20秒
                //警笛
                CMD_LocFunction(ADDR_DCC + 3, 11, 0);
                CMD_Wait(2000);
                //定速時間
                CMD_Wait(20000);    //定速時間20秒

                //繰り返しの例。車両のスピードを変える(減速1）
                for( int i = 170; i > 32; i--)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

                //絶気モード
                //CMD_LocFunction(ADDR_DCC + 3, 3, 1);
                //CMD_Wait(2000);
                
                //繰り返し。車両のスピードを変える(減速2）
                for( int i = 32; i > 0; i--)
                   {
                    CMD_LocSpeed(ADDR_DCC + 3, i * 4 );
                    CMD_Wait(100);
                   }

                //停止時間
                CMD_LocSpeed(ADDR_DCC + 3, 0);
                CMD_Wait(2000);
                
                //絶気モード解除
                //CMD_LocFunction(ADDR_DCC + 3, 3, 0);
                //CMD_Wait(2000);
                
                CMD_Wait(40000);   //停止時間40秒

	}

	if( digitalRead(11) == 0)
	{
		Serial.println("Button 2");
		
		sButtonState = 2;
		CMD_Wait(100);
		//急停車
		CMD_LocSpeed(ADDR_DCC + 3, 0);
	}
	
	if( digitalRead(12) == 0)
	{
		Serial.println("Button 3");
		
		sButtonState = 3;
		CMD_Wait(100);
	}
	
	/* S88のスキャン */
	CMD_ScanS88();
	
	
	if( CMD_GetS88(0) == 1)
	{
		Serial.println("S88 Detected");
		
		CMD_LocSpeed(ADDR_DCC + 3, 100);
		CMD_Wait(500);
		CMD_LocSpeed(ADDR_DCC + 3, 75);
		CMD_Wait(500);
		CMD_LocSpeed(ADDR_DCC + 3, 50);
		CMD_Wait(500);
		CMD_LocSpeed(ADDR_DCC + 3, 25);
		CMD_Wait(500);
		CMD_LocSpeed(ADDR_DCC + 3, 0);
		CMD_Wait(1000);
		
		//REV
		CMD_LocDirection(ADDR_DCC + 3, 1);
		Serial.println("REV");
		
		//警笛
		CMD_LocFunction(ADDR_DCC + 3, 2, 1);
		CMD_Wait(2000);
		CMD_LocFunction(ADDR_DCC + 3, 2, 0);

		/* 加速 */
		CMD_LocSpeed(ADDR_DCC + 3, 25);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 50);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 75);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 100);
		CMD_Wait(3000);
		
		CMD_LocFunction(ADDR_DCC + 3, 3, 1);
		CMD_Wait(2000);
		CMD_LocFunction(ADDR_DCC + 3, 3, 0);
		
		/* 減速 */
		CMD_LocSpeed(ADDR_DCC + 3, 75);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 50);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 25);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 0);
		CMD_Wait(2000);		
		
		//FWD
		CMD_LocDirection(ADDR_DCC + 3, 0);
		Serial.println("FWD");
		
		//Clear
		CMD_ScanS88();
		
		//警笛
		CMD_LocFunction(ADDR_DCC + 3, 2, 1);
		CMD_Wait(100);
		CMD_LocFunction(ADDR_DCC + 3, 2, 0);
		
		/* 加速 */
		CMD_LocSpeed(ADDR_DCC + 3, 25);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 50);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 75);
		CMD_Wait(1000);
		CMD_LocSpeed(ADDR_DCC + 3, 100);
	}
}

